@@ -1273,7 +1273,7 @@ class ByteCodeMachine extends StackMachine {
     private void opEndLineSb()  {
         if (s == end) {
             if (Config.USE_NEWLINE_AT_END_OF_STRING_HAS_EMPTY_LINE) {
-                if (str == end || bytes[sprev] != Encoding.NEW_LINE) {
+                if (str == end || !(sprev < end && bytes[sprev] == Encoding.NEW_LINE)) {
                     if (isNotEol(msaOptions)) opFail();
                 }
                 return;
@@ -1320,6 +1320,7 @@ class ByteCodeMachine extends StackMachine {
     private void opMemoryStart() {
         int mem = code[ip++];
         repeatStk[memStartStk + mem] = s;
+        repeatStk[memEndStk + mem] = -1;
     }
 
     private void opMemoryEndPush() {
