@@ -41,10 +41,12 @@ public interface Config extends org.jcodings.Config {
     final boolean USE_CAPTURE_HISTORY = false;
     final boolean USE_VARIABLE_META_CHARS = true;
     final boolean USE_WORD_BEGIN_END = true;                                /* "\<": word-begin, "\>": word-end */
-    final boolean USE_POSIX_API_REGION_OPTION = true;                           /* needed for POSIX API support */
     final boolean USE_FIND_LONGEST_SEARCH_ALL_OF_RANGE = true;
+    final boolean USE_SUNDAY_QUICK_SEARCH = true;
     final boolean USE_CEC = false;
     final boolean USE_DYNAMIC_OPTION = false;
+    final boolean USE_BYTE_MAP = OptExactInfo.OPT_EXACT_MAXLEN <= CHAR_TABLE_SIZE;
+    final boolean USE_INT_MAP_BACKWARD = false;
 
     final int NREGION                   = 10;
     final int MAX_BACKREF_NUM           = 1000;
@@ -52,19 +54,11 @@ public interface Config extends org.jcodings.Config {
     final int MAX_REPEAT_NUM            = 100000;
     final int MAX_MULTI_BYTE_RANGES_NUM = 10000;
 
-
-    final boolean USE_WARN = true;
-
     // internal config
-    final boolean USE_PARSE_TREE_NODE_RECYCLE       = true;
     final boolean USE_OP_PUSH_OR_JUMP_EXACT         = true;
     final boolean USE_QTFR_PEEK_NEXT                = true;
 
     final int INIT_MATCH_STACK_SIZE                 = 64;
-    final int DEFAULT_MATCH_STACK_LIMIT_SIZE        = 0;        /* unlimited */
-    final int NUMBER_OF_POOLED_STACKS               = 4;
-
-
 
     final boolean DONT_OPTIMIZE                     = false;
 
