@@ -26,7 +26,7 @@ import org.jcodings.Encoding;
 import org.jcodings.IntHolder;
 import org.jcodings.constants.CharacterType;
 import org.jcodings.specific.ASCIIEncoding;
-import org.joni.constants.AnchorType;
+import org.joni.constants.internal.AnchorType;
 
 public abstract class Matcher extends IntHolder {
     public static final int FAILED = -1;
@@ -48,22 +48,23 @@ public abstract class Matcher extends IntHolder {
     protected int msaBegin;
     protected int msaEnd;
 
-    Matcher(Regex regex, byte[]bytes, int p, int end) {
+    Matcher(Regex regex, Region region, byte[]bytes, int p, int end) {
         this.regex = regex;
         this.enc = regex.enc;
         this.bytes = bytes;
         this.str = p;
         this.end = end;
-
-        this.msaRegion = regex.numMem == 0 ? null : new Region(regex.numMem + 1);
+        this.msaRegion = region;
     }
 
     // main matching method
-    protected abstract int matchAt(int range, int sstart, int sprev) throws InterruptedException;
+    protected abstract int matchAt(int range, int sstart, int sprev, boolean interrupt) throws InterruptedException;
 
     protected abstract void stateCheckBuffInit(int strLength, int offset, int stateNum);
     protected abstract void stateCheckBuffClear();
 
+    public abstract void interrupt();
+
     public final Region getRegion() {
         return msaRegion;
     }
@@ -88,13 +89,17 @@ public abstract class Matcher extends IntHolder {
 
     public final int match(int at, int range, int option) {
         try {
-            return matchInterruptible(at, range, option);
+            return matchCommon(at, range, option, false);
         } catch (InterruptedException ex) {
             return INTERRUPTED;
         }
     }
 
     public final int matchInterruptible(int at, int range, int option) throws InterruptedException {
+        return matchCommon(at, range, option, true);
+    }
+
+    private final int matchCommon(int at, int range, int option, boolean interrupt) throws InterruptedException {
         msaInit(option, at);
 
         if (Config.USE_CEC) {
@@ -105,9 +110,9 @@ public abstract class Matcher extends IntHolder {
         int prev = enc.prevCharHead(bytes, str, at, end);
 
         if (Config.USE_MATCH_RANGE_MUST_BE_INSIDE_OF_SPECIFIED_RANGE) {
-            return matchAt(end /*range*/, at, prev);
+            return matchAt(end /*range*/, at, prev, interrupt);
         } else {
-            return matchAt(range /*range*/, at, prev);
+            return matchAt(range /*range*/, at, prev, interrupt);
         }
     }
 
@@ -128,7 +133,8 @@ public abstract class Matcher extends IntHolder {
         }
 
         retry:while (true) {
-            p = regex.searchAlgorithm.search(this, bytes, p, end, range);
+            if (Config.DEBUG_SEARCH) debugSearch(regex.forward.getName(), p, end, range);
+            p = regex.forward.search(this, bytes, p, end, range);
 
             if (p != -1 && p < range) {
                 if (p - regex.dMin < s) {
@@ -214,7 +220,7 @@ public abstract class Matcher extends IntHolder {
         int p = s;
 
         retry:while (true) {
-            p = regex.searchAlgorithm.searchBackward(this, bytes, range, adjrange, end, p, s, range);
+            p = regex.backward.search(this, bytes, range, adjrange, end, p, s, range);
 
             if (p != -1) {
                 if (regex.subAnchor != 0) {
@@ -265,26 +271,26 @@ public abstract class Matcher extends IntHolder {
     }
 
     // MATCH_AND_RETURN_CHECK
-    private boolean matchCheck(int upperRange, int s, int prev) throws InterruptedException {
+    private boolean matchCheck(int upperRange, int s, int prev, boolean interrupt) throws InterruptedException {
         if (Config.USE_MATCH_RANGE_MUST_BE_INSIDE_OF_SPECIFIED_RANGE) {
             if (Config.USE_FIND_LONGEST_SEARCH_ALL_OF_RANGE) {
                 //range = upperRange;
-                if (matchAt(upperRange, s, prev) != -1) {
+                if (matchAt(upperRange, s, prev, interrupt) != -1) {
                     if (!isFindLongest(regex.options)) return true;
                 }
             } else {
                 //range = upperRange;
-                if (matchAt(upperRange, s, prev) != -1) return true;
+                if (matchAt(upperRange, s, prev, interrupt) != -1) return true;
             }
         } else {
             if (Config.USE_FIND_LONGEST_SEARCH_ALL_OF_RANGE) {
-                if (matchAt(end, s, prev) != -1) {
+                if (matchAt(end, s, prev, interrupt) != -1) {
                     //range = upperRange;
                     if (!isFindLongest(regex.options)) return true;
                 }
             } else {
                 //range = upperRange;
-                if (matchAt(end, s, prev) != -1) return true;
+                if (matchAt(end, s, prev, interrupt) != -1) return true;
             }
         }
         return false;
@@ -292,13 +298,17 @@ public abstract class Matcher extends IntHolder {
 
     public final int search(int start, int range, int option) {
         try {
-            return searchInterruptible(start, range, option);
+            return searchCommon(start, range, option, false);
         } catch (InterruptedException ex) {
             return INTERRUPTED;
         }
     }
 
     public final int searchInterruptible(int start, int range, int option) throws InterruptedException {
+        return searchCommon(start, range, option, true);
+    }
+
+    private final int searchCommon(int start, int range, int option, boolean interrupt) throws InterruptedException {
         int s, prev;
         int origStart = start;
         int origRange = range;
@@ -376,7 +386,7 @@ public abstract class Matcher extends IntHolder {
 
                 if (Config.USE_CEC) stateCheckBuffClear();
 
-                if (matchCheck(end, s, prev)) return match(s);
+                if (matchCheck(end, s, prev, interrupt)) return match(s);
                 return mismatch();
             }
             return FAILED; // goto mismatch_no_msa;
@@ -398,7 +408,7 @@ public abstract class Matcher extends IntHolder {
                 prev = 0; // -1
             }
 
-            if (regex.searchAlgorithm != SearchAlgorithm.NONE) {
+            if (regex.forward != null) {
                 int schRange = range;
                 if (regex.dMax != 0) {
                     if (regex.dMax == MinMaxLen.INFINITE_DISTANCE) {
@@ -418,7 +428,7 @@ public abstract class Matcher extends IntHolder {
                             prev = value;
                         }
                         while (s <= high) {
-                            if (matchCheck(origRange, s, prev)) return match(s); // ???
+                            if (matchCheck(origRange, s, prev, interrupt)) return match(s); // ???
                             prev = s;
                             s += enc.length(bytes, s, end);
                         }
@@ -430,7 +440,7 @@ public abstract class Matcher extends IntHolder {
 
                     if ((regex.anchor & AnchorType.ANYCHAR_STAR) != 0) {
                         do {
-                            if (matchCheck(origRange, s, prev)) return match(s);
+                            if (matchCheck(origRange, s, prev, interrupt)) return match(s);
                             prev = s;
                             s += enc.length(bytes, s, end);
 
@@ -447,13 +457,13 @@ public abstract class Matcher extends IntHolder {
             }
 
             do {
-                if (matchCheck(origRange, s, prev)) return match(s);
+                if (matchCheck(origRange, s, prev, interrupt)) return match(s);
                 prev = s;
                 s += enc.length(bytes, s, end);
             } while (s < range);
 
             if (s == range) { /* because empty match with /$/. */
-                if (matchCheck(origRange, s, prev)) return match(s);
+                if (matchCheck(origRange, s, prev, interrupt)) return match(s);
             }
         } else { /* backward search */
             if (Config.USE_MATCH_RANGE_MUST_BE_INSIDE_OF_SPECIFIED_RANGE) {
@@ -462,7 +472,7 @@ public abstract class Matcher extends IntHolder {
                 }
             }
 
-            if (regex.searchAlgorithm != SearchAlgorithm.NONE) {
+            if (regex.backward != null) {
                 int adjrange;
                 if (range < end) {
                     adjrange = enc.leftAdjustCharHead(bytes, str, range, end);
@@ -477,7 +487,7 @@ public abstract class Matcher extends IntHolder {
                         if (s > high) s = high;
                         while (s != -1 && s >= low) {
                             prev = enc.prevCharHead(bytes, str, s, end);
-                            if (matchCheck(origStart, s, prev)) return match(s);
+                            if (matchCheck(origStart, s, prev, interrupt)) return match(s);
                             s = prev;
                         }
                     } while (s >= range);
@@ -504,7 +514,7 @@ public abstract class Matcher extends IntHolder {
 
             do {
                 prev = enc.prevCharHead(bytes, str, s, end);
-                if (matchCheck(origStart, s, prev)) return match(s);
+                if (matchCheck(origStart, s, prev, interrupt)) return match(s);
                 s = prev;
             } while (s >= range);
 
@@ -603,4 +613,9 @@ public abstract class Matcher extends IntHolder {
                 ", high: " + (high - str));
         }
     }
+
+    static void debugSearch(String name, int textP, int textEnd, int textRange) {
+        Config.log.println(name + ": text: " + textP + ", text_end: " + textEnd + ", text_range: " + textRange);
+    }
+
 }
