@@ -19,7 +19,7 @@
  */
 package org.joni;
 
-final class StackEntry {
+class StackEntry {
     int type;
     private int E1, E2, E3, E4;
 
@@ -46,10 +46,10 @@ final class StackEntry {
         return E3;
     }
 
-    void setStateCheck(int check) {
-        E4 = check;
+    void setPKeep(int pkeep) {
+        E4 = pkeep;
     }
-    int getStateCheck() {
+    int getPKeep() {
         return E4;
     }
 
@@ -161,4 +161,31 @@ final class StackEntry {
     int getCallFramePStr() {
         return E3;
     }
+
+    /* absent position */
+    void setAbsentStr(int pos) {
+        E1 = pos;
+    }
+    int getAbsentStr() {
+        return E1;
+    }
+
+    void setAbsentEndStr(int pos) {
+        E2 = pos;
+    }
+    int getAbsentEndStr() {
+        return E2;
+    }
 }
+
+final class SCStackEntry extends StackEntry {
+    private int E5;
+
+    void setStateCheck(int check) {
+        E5 = check;
+    }
+
+    int getStateCheck() {
+        return E5;
+    }
+}
\ No newline at end of file
